#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "tscp.h"


FQLink::FQLink(Network *n, int maxBytes, int pktSize,
	       float b, float d,
	       int max_flows, int max_flow_id, int monitor)
{
    int max_msgs = maxBytes/pktSize;
    segmentSize = pktSize;
    bandwidth = b;
    delay = d;
    numMessages = 0;
    maxMessages = max_msgs;
    maxFlows = max_flows;
    flows = new FQFlow[maxFlows];

    for (int i=0; i<maxFlows; i++)
	flows[i].InitFlow(max_msgs, segmentSize);
    
    maxFlowId = max_flow_id;
    invFlowIdMap = new int[maxFlowId];
    for (int j=0; j<maxFlowId; j++)
	invFlowIdMap[j] = -1;
    
    net = n;
    linkId = n->GetLinkNo();
    virtRound = 0;
    virtTime = 0;
    numActiveFlows = 0;
    numRegisteredFlows = 0;
    sizeHeap = new Heap(maxFlows);
    priorHeap = new Heap(maxFlows);
    stillActiveHeap = new Heap(maxFlows);

    statsNumBytes = 0;
    statsNumDrops = 0;
    statsNumPackets = 0;
    statsQueueSize = 0;
    statsLastEvent = 0;
    n->SetLink(linkId, this);

    if (monitor)
	statsObj = new StatsObject(100000);
    else
	statsObj = NULL;
}

void FQLink::PrintStatus()
{
    printf("**********LINK %d*********\n", linkId);
    printf("Flows: %d, Messages: %d\n", numActiveFlows, numMessages);
    int msgs = 0;
    for (int i=0; i<maxFlows; i++)
	if (!flows[i].IsDormant() && flows[i].NumMessages()) {
	    msgs += flows[i].NumMessages();
	    flows[i].PrintStatus();
	}
    if (numMessages != msgs)
	assert(0);
}

void FQLink::RegisterFlow(int flowId)
{
    assert(invFlowIdMap[flowId] == -1);
    numRegisteredFlows++;
    for (int i=0; i<maxFlows; i++) {
	if (flows[i].IsDormant()) {
	    flows[i].StartFlow(flowId);
	    invFlowIdMap[flowId] = i;
	    flows[i].AddToSizeHeap(sizeHeap);
	    return;
	}
    }
    assert(0);
}

void FQLink::UnregisterFlow(int flowId)
{
    int num = invFlowIdMap[flowId];
    if ((flows[num].NumMessages() > 0) || flows[num].InActiveHeap())
	numActiveFlows--;
    numMessages -= flows[num].NumMessages();
    flows[num].EndFlow(sizeHeap, priorHeap, stillActiveHeap, &xmitFlow);
    invFlowIdMap[flowId] = -1;
    numRegisteredFlows--;    
}

void FQLink::UpdateVirtualTime(float currTime)
{
    float newRound;
    while (1) {
	float nextFinRound;
	if (numActiveFlows == 0) {
	    newRound = virtRound;
	    break;
	}
	newRound = virtRound + (currTime - virtTime)*bandwidth/numActiveFlows;
	if (stillActiveHeap->IsEmpty()) break;
	nextFinRound = stillActiveHeap->PeekMinCost();
	if (nextFinRound > newRound) break;
	FQFlow *f = (FQFlow *)stillActiveHeap->DeleteMin();
	f->UnsetActive();
	virtTime = virtTime + (nextFinRound - virtRound)*numActiveFlows/bandwidth;
	virtRound = nextFinRound;
	if (f->NumMessages() == 0)
	    numActiveFlows--;
    }
    virtRound = newRound;
    virtTime = currTime;
#ifdef DEBUG
    printf("#############curr: %f, virt: %f, round: %f, peek: %f\n",
	   currTime, virtTime, virtRound,
	   (stillActiveHeap->IsEmpty() ? 0 : stillActiveHeap->PeekMinCost()));
#endif
    return;
}

void FQLink::UpdateQueueStats(float curr)
{
    statsQueueSize += (curr - statsLastEvent)*numMessages;
    statsLastEvent = curr;
}

void FQLink::EnableStats(float curr)
{
    statsNumBytes = 0;
    statsNumDrops = 0;
    statsNumPackets = 0;    
    statsQueueSize = 0;
    statsLastEvent = curr;
    statsInitTime = curr;
    statsMon = 1;
}

void FQLink::PrintStats(float curr)
{
    if (statsObj == NULL) {
	printf("%f %f %f\n", (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    
	printf("Link %d stats -- Bytes: %d, Drops: %d, Drop-per: %f, Thruput: %f kb/s, Queue: %f\n",
	       linkId, statsNumBytes, statsNumDrops,
	       (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    } else {
#if 0
	printf("%f %f %f %f\n", (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime),
	       statsObj->QueryPercentile(95));
#endif
	printf("Drop-per: %f, Thruput: %f kb/s, Queue: %f\n",
	       (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    }
}

void FQLink::HandleMessage(Message *m, float currTime)
{
    int num = invFlowIdMap[m->flowNo];

    statsNumPackets++;
#ifdef DEBUG
    printf("%f: message %d:%d enqueued on link %d\n", currTime, m->flowNo, m->seqno, linkId);
#endif
    assert(m->size <= segmentSize);
    
    /* straggler message, just ignore it */
    if (num == -1)
	return;
    
    if (!flows[num].AddTail(m, sizeHeap)) {
	delete m;
	/*	printf("%f: DROPPED %d:%d\n", currTime, m->flowNo, m->seqno); */
	statsNumDrops++;
	return;
    }

    if (flows[num].NumMessages() == 1) {
	flows[num].SetStartRound(virtRound, priorHeap);
	if  (!flows[num].InActiveHeap())
	    numActiveFlows++;
    }

    UpdateQueueStats(currTime);
    numMessages++;
    
    if (numMessages > maxMessages) {
	FQFlow *f = (FQFlow *)sizeHeap->DeleteMin();
	Message *m = f->RemoveTail();
	/*	printf("%f: DROPPED %d:%d\n", currTime, m->flowNo, m->seqno); */
	statsNumDrops++;	
	f->AddToSizeHeap(sizeHeap);
	numMessages--;
    }
    
    /* if only message in queue, schedule its departure */
    if (numMessages == 1) {
	float del = m->size/bandwidth;
	xmitFlow = (FQFlow *)priorHeap->DeleteMin();
	Event *e = new Event(currTime + del, this, FIRE_TIMER, NULL, currTime);
	net->AddEvent(e);
    }
}

#if 0
void FQLink::CheckFairness(float curr)
{
    float lastSeen = virtRound + 1;
    FQFlow *lastFlow = NULL;
    for (int i=0; i<maxFlows; i++) {
	if (flows[i].startRound < lastSeen && flows[i].NumMessages() > 0) {
	    lastSeen = flows[i].lastService;
	    lastFlow = &flows[i];
	}
    }

    if (lastFlow->lastService < xmitFlow->lastService)
	printf("problem\n");
}
#endif

#if 0
float FQLink::ReallocBWShare(FQFlow *f)
{
    float oldShare = f->GetBWShare();
    float newShare = bandwidth/numActiveFlows;

    if (newShare < oldShare) {
	reclaimedBW += (oldShare - newShare);
	f->SetBWShare(newShare);
	return newShare;
    } else {
	if (reclaimedBW < (newShare - oldShare)) {
	    reclaimedBW = 0;
	    f->SetBWShare(reclaimedBW + oldShare);
	    return reclaimedBW + oldShare;
	} else {
	    reclaimedBW -= (newShare - oldShare);
	    f->SetBWShare(newShare);
	    return newShare;
	}
    }
}
#endif

void FQLink::HandleEvent(Event *ev)
{
    Message *m;
    float curr = ev->time;
    Event *e;

    UpdateVirtualTime(curr);
    if (ev->evType == ENQUEUE_TIMER) {
	HandleMessage((Message *)(ev->data), curr);
    } else if (ev->evType == FIRE_TIMER) {
	if (numMessages == 0)
	    return;

	if (statsMon && statsObj)
	    statsObj->RecordObservation(numMessages);	    

	if (xmitFlow && xmitFlow->NumMessages() > 0) {
	    m = xmitFlow->RemoveHead(priorHeap, sizeHeap);
	    xmitFlow->AddToActiveHeap(stillActiveHeap);
	    UpdateQueueStats(curr);
	    statsNumBytes += m->size;
	    numMessages--;

	    int i;
	    float qszFactor = xmitFlow->NumMessages()/((float)maxMessages/numRegisteredFlows);
	    float myshare;

#if 0
	    if (qszFactor > 0.33)
		myshare = bandwidth/2*numRegisteredFlows;
	    else
		myshare = bandwidth/numActiveFlows;
#endif

	    myshare = bandwidth/numRegisteredFlows;
	    m->bwshare = (m->bwshare < myshare ? m->bwshare : myshare);
	    m->qlen = (m->qlen < qszFactor ? qszFactor : m->qlen);
	    /* schedule the handling of the current message by the next agent */
	    for (i=0; i<m->conn->numAgents; i++)
		if (m->conn->agents[i] == this)
		    break;
	    assert(i != m->conn->numAgents);
	    e = new Event(curr + delay, m->conn->agents[i+1], ENQUEUE_TIMER, m, curr);
	    net->AddEvent(e);
	}

	if (numMessages != 0) {
	    float del;
	    Message *nextmsg;

	    xmitFlow = (FQFlow *)priorHeap->DeleteMin();

	    /* schedule the next departure */
	    nextmsg = xmitFlow->PeepHead();
	    del = nextmsg->size/bandwidth;
	    e = new Event(curr + del, this, FIRE_TIMER, NULL, curr);
	    net->AddEvent(e);
	}

    } else
	assert(0);
    UpdateVirtualTime(curr);
    /* PrintStatus(); */
}

/*{for 1 = 1 : 100,
	   i love u arvind
	   }*/

