#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "tscp.h"

/* #define DEBUG*/

FifoLink::FifoLink(Network *n, int bufferSize, float b, float d, int monitor)
{
    numMessages = 0;
    numBurstBytes = 0;
    numBytes = 0;
    maxBytes = bufferSize;
    begin = end = 0;
    bandwidth = b;
    delay = d;
    maxMessages = bufferSize/100;
    buffer = new (Message *)[maxMessages];
    net = n;
    linkId = n->GetLinkNo();
    n->SetLink(linkId, this);
    if (monitor)
	statsObj = new StatsObject(100000);
    else
	statsObj = NULL;
}

void FifoLink::UpdateQueueStats(float curr)
{
    /*    statsQueueSize += (curr - statsLastEvent)*numMessages; */
    statsQueueSize += (curr - statsLastEvent)*(numBytes/1250.0);
    statsBurstQueueSize += (curr - statsLastEvent)*(numBurstBytes/1250.0);
    statsLastEvent = curr;
}

void FifoLink::EnableStats(float curr)
{
    statsNumBytes = 0;
    statsNumDrops = 0;
    statsNumPackets++;
    statsQueueSize = 0;
    statsBurstQueueSize = 0;
    statsLastEvent = curr;
    statsInitTime = curr;
    statsMon = 1;
}

void FifoLink::PrintStats(float curr)
{
    if (statsObj == NULL) {
	printf("%f %f %f\n", (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    
	printf("Link %d stats -- Bytes: %d, Drops: %d, Drop-per: %f, Thruput: %f kb/s, Queue: %f\n",
	       linkId, statsNumBytes, statsNumDrops,
	       (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    } else {
#if 0
	printf("%f %f %f %f\n", (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime),
	       
	       statsObj->QueryPercentile(95));
#endif
	printf("Drop-per: %f, Thruput: %f kb/s, Queue: %f\n",
	       (100.0*statsNumDrops)/statsNumPackets,
	       8.0*statsNumBytes/(curr - statsInitTime),
	       statsQueueSize/(curr - statsInitTime));
    }
}

void FifoLink::HandleMessage(Message *m, float currTime)
{

    statsNumPackets++;
    if (numMessages == maxMessages)
	assert(0);
    
    /* enqueue the message in the buffer if there is space */
    if ((numBytes + m->size) > maxBytes) {
	/*	printf("%f: DROPPED %d:%d on %d\n", currTime, m->flowNo, m->seqno, linkId); */
	delete m;
	statsNumDrops++;
	return;
    }

    buffer[end] = m;
    end = (end + 1) % (maxMessages);
    numMessages++;
    if (m->burstMsg)
	numBurstBytes += m->size;
    numBytes += m->size;
    UpdateQueueStats(currTime);
#ifdef DEBUG
    if (statsObj)
	printf("%f: message %d:%d enqueued on link %d\n", currTime, m->flowNo, m->seqno, linkId);	
    printf("%f: message %d:%d enqueued on link %d\n", currTime, m->flowNo, m->seqno, linkId);
#endif
    /* if only message in queue, schedule its departure */
    if (numMessages == 1) {
	float del = m->size/bandwidth;
	Event *e = new Event(currTime + del, this, FIRE_TIMER, NULL, currTime);
	net->AddEvent(e);
    }
}

void FifoLink::HandleEvent(Event *ev)
{
    Message *m;
    float curr = ev->time;
    Event *e;

    if (ev->evType == ENQUEUE_TIMER) {
	HandleMessage((Message *)(ev->data), curr);
    } else if (ev->evType == FIRE_TIMER) {
	if (numMessages == 0)
	    return;

	if (statsMon && statsObj)
	    statsObj->RecordObservation(numBytes/1250);
	m = buffer[begin];
	statsNumBytes += m->size;
	numMessages--;
	numBytes -= m->size;
	if (m->burstMsg)
	    numBurstBytes -= m->size;
	UpdateQueueStats(curr);
#ifdef DEBUG
	if (statsObj)
	    printf("%f: message %d:%d of size %d xmitted by link %d\n", curr, m->flowNo, m->seqno, m->size, linkId);
	printf("%f: message %d:%d of size %d xmitted by link %d\n", curr, m->flowNo, m->seqno, m->size, linkId);
#endif

	if (numMessages == 0) {
	    begin = 0;
	    end = 0;
	} else {
	    float del;
	    Message *nextmsg;

	    /* schedule the next departure */
	    begin = (begin + 1) % (maxMessages);
	    nextmsg = buffer[begin];
	    del = nextmsg->size/bandwidth;
	    e = new Event(curr + del, this, FIRE_TIMER, NULL, curr);
	    net->AddEvent(e);
	}

	{
	    int i;

	    /* schedule the handling of the current message by the next agent */
	    for (i=0; i<m->conn->numAgents; i++)
		if (m->conn->agents[i] == this)
		    break;
	    assert(i != m->conn->numAgents);
	    e = new Event(curr + delay, m->conn->agents[i+1], ENQUEUE_TIMER, m, curr);
	    net->AddEvent(e);
	}
    } else
	assert(0);
}
