#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "tscp.h"

GenericReceiver::GenericReceiver(Network *n)
{
    net = n;
    conn = NULL;
}

GenericReceiver::GenericReceiver(Network *n, Connection *c)
{
    net = n;
    conn = c;
}

void GenericReceiver::HandleMessage(Message *m, float curr)
{
#ifdef DEBUG
    printf("%f: message %d:%d received\n", curr, m->flowNo, m->seqno);
#endif
    m->size = 40;
    m->conn = conn;
    m->conn->agents[0]->HandleMessage(m, curr);
}

void GenericReceiver::HandleEvent(Event *e)
{
    if (e->evType == ENQUEUE_TIMER) {
	HandleMessage((Message *)(e->data), e->time);
    }    
}

void GenericReceiver::RegisterFlow(int flowNo)
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->RegisterFlow(flowNo);
    }
}

void GenericReceiver::UnregisterFlow(int flowNo)
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->UnregisterFlow(flowNo);
    }
}

TcpReceiver::TcpReceiver(Network *n) : GenericReceiver(n)
{
    rememberedAck = -1;
    maxWindow = 800;
    acked = new int [maxWindow];
    begin = 0;
    nextAck = 0;
    numOutOfOrderAcks = 0;
    for (int i=0; i<maxWindow; i++)
	acked[i] = 0;
}

TcpReceiver::TcpReceiver(Network *n, Connection *c) : GenericReceiver(n, c)
{
    rememberedAck = -1;
    maxWindow = 800;
    acked = new int [maxWindow];
    begin = 0;
    nextAck = 0;
    numOutOfOrderAcks = 0;    
    for (int i=0; i<maxWindow; i++)
	acked[i] = 0;
}

void TcpReceiver::HandleMessage(Message *m, float curr)
{
#ifdef DEBUG
    printf("%f: message %d:%d received\n", curr, m->flowNo, m->seqno);
#endif
    int seq = m->seqno;
    int suppressAck = 0;

    if (seq < nextAck) {
	delete m;
	return;
    }

    if ((numOutOfOrderAcks == 0)
	&& (seq == nextAck) && (rememberedAck == -1)
	&& (seq != 0) && (seq != 1)
	&& (m->fin != 1))
	suppressAck = 1;

    if (seq != nextAck) {
	acked[(begin + (seq - nextAck) + maxWindow) % maxWindow] = 1;
	numOutOfOrderAcks++;
    } else {
	nextAck = seq + 1;
	begin = (begin + 1) % maxWindow;
	while (acked[begin]) {
	    nextAck = nextAck+1;
	    acked[begin] = 0;
	    begin = (begin + 1) % maxWindow;
	    numOutOfOrderAcks--;
	}
    }

    if (!suppressAck) {
	m->size = 40;
	m->conn = conn;
	m->ack2 = rememberedAck;
	m->conn->agents[0]->HandleMessage(m, curr);
	rememberedAck = -1;
    } else {
	rememberedAck = seq;
	Event *ev = new Event(curr + 200, this, DELACK_TIMER, m, curr);
	net->AddEvent(ev);
    }
}

void TcpReceiver::HandleEvent(Event *e)
{
    if (e->evType == ENQUEUE_TIMER) {
	HandleMessage((Message *)(e->data), e->time);
    } else if (e->evType == DELACK_TIMER) {
	Message *m = (Message *)(e->data);
	if (rememberedAck == -1) {
	    delete m;
	} else {
	    m->size = 40;
	    m->conn = conn;
	    m->conn->agents[0]->HandleMessage(m, e->time);
	}
    }
}
