#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "samples.h"


Samples::Samples(int max, float hor)
{
    samples = new float [max];
    scratch = new float [max];
    times = new float [max];
    maxSamples = max;
    end = 0;
    for (int i=0; i<max; i++)
	times[i] = -hor;
    horizon = hor;
}

void Samples::Clean()
{
    end = 0;
    for (int i=0; i<maxSamples; i++)
	times[i] = -horizon;
}

void Samples::RecordSample(float newTime, float newVal)
{
    assert(times[end] <= newTime - horizon);
    times[end] = newTime;
    samples[end] = newVal;
    end = ((end + 1) % maxSamples);
}


int Samples::LookupIndex(float targTime)
{
    int i=(end-1+maxSamples) % maxSamples;
    while (times[i] > targTime)
	i = (i-1+maxSamples) % maxSamples;
    return i;
}

float Samples::LookupSample(float targTime)
{
    return samples[LookupIndex(targTime)];
}

float Samples::SumSamples(float t1, float t2)
{
    int fin = LookupIndex(t2);
    int beg = LookupIndex(t1);
    float cum = 0;

    int i = beg;
    fin = (fin + 1) % maxSamples;
    do {
	cum += samples[i];
	i = (i + 1) % maxSamples;
    } while (i != fin);
    return cum;
}

float Samples::AverageSample(float t1, float t2)
{
    int fin = LookupIndex(t2);
    int beg = LookupIndex(t1);
    float cum = 0;
    float cumTime = 0;

    int i = beg;
    
    float beginTime = t1;
    float endTime = times[(i + 1) % maxSamples];
    while (i != fin) {
	cum += samples[i]*(endTime - beginTime);
	cumTime += (endTime - beginTime);
	i = (i + 1) % maxSamples;
	beginTime = endTime;
	endTime = times[(i + 1) % maxSamples];
    }
    
    cum += samples[i]*(t2 - beginTime);
    cumTime += (t2 - beginTime);
    assert(fabs(cumTime - (t2 - t1)) < 0.01);
    return cum/(t2 - t1);
}


float Samples::VarianceSample(float t1, float t2)
{
    int fin = LookupIndex(t2);
    int beg = LookupIndex(t1);
    float cum = 0;
    int i, num;

    fin = (fin + 1) % maxSamples;
    for (i = beg, num = 0; i != fin; i = (i + 1) % maxSamples) {
	cum += samples[i];
	num++;
    }

    float avg = cum/num;
    for (i = beg; i != fin; i = (i + 1) % maxSamples) {
	cum += (samples[i] - avg)*(samples[i] - avg);
    }
    return sqrt(cum/num);
}

static int floatcompare(const void *e1, const void *e2)
{
    float f1 = *(float *)e1;
    float f2 = *(float *)e2;
    if (f1 < f2)
	return -1;
    else if (f1 == f2)
	return 0;
    else
	return 1;
}


float Samples::NinetyPercentSample(float t1, float t2)
{
    int fin = LookupIndex(t2);
    int beg = LookupIndex(t1);
    int num;

    if (beg <= fin)
	num = fin - beg + 1;
    else
	num = fin + maxSamples - beg + 1;

    for (int i=0; i<num; i++)
	scratch[i] = samples[(beg + i)%maxSamples];

    qsort(scratch, num, sizeof(float), floatcompare);
    return scratch[(int)(0.9*num)];
}

float Samples::PercentileSample(float t1, float t2, int perc)
{
    int fin = LookupIndex(t2);
    int beg = LookupIndex(t1);
    int num;

    if (beg <= fin)
	num = fin - beg + 1;
    else
	num = fin + maxSamples - beg + 1;

    for (int i=0; i<num; i++)
	scratch[i] = samples[(beg + i)%maxSamples];

    qsort(scratch, num, sizeof(float), floatcompare);
    if ((perc*num/100) == num)
	return scratch[num-1];
    else
	return scratch[perc*num/100];
}

#if 0
int main(int argc, char **argv)
{
    Samples s(100, 10);
    for (int i=0; i<150; i++)
	s.RecordSample(0.2*i, i);
    printf("%f %f %f %f %f\n", s.LookupSample(18.1), s.LookupSample(21.05),
	   s.SumSamples(18.1, 21.05), s.AverageSample(18.1, 21.05),
	   s.NinetyPercentSample(18.1, 18.15));
}
#endif
