#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "tscp.h"

AckScoreboard::AckScoreboard(int maxW)
{
    nextack = 0;
    acked = new int[maxW];
    ackedTime = new float[maxW];    
    sent = new int[maxW];
    resent = new int[maxW];    
    msgSize = new int[maxW];
    fin = new int[maxW];
    times = new float[maxW];
    maxWindow = maxW;
    maxBytesWindow = maxW*1600;
    begin = 0;
    nextseqno = 0;
    numNewAcks = 0;
    for (int i=0; i<maxW; i++) {
	sent[i] = 0;
	resent[i] = 0;
	acked[i] = 0;
    }
    bytesInTransit = 0;
    segmentsInTransit = 0;
}

void AckScoreboard::CleanScoreboard()
{
    nextack = 0;
    nextseqno = 0;    
    begin = 0;
    numNewAcks = 0;
    for (int i=0; i<maxWindow; i++) {
	sent[i] = 0;
	resent[i] = 0;
	acked[i] = 0;
    }
    bytesInTransit = 0;
    segmentsInTransit = 0;
}

AckScoreboard::AckScoreboard(int maxW, int bytesWindow)
{
    nextack = 0;
    acked = new int[maxW];
    resent = new int[maxW];        
    sent = new int[maxW];    
    msgSize = new int[maxW];
    fin = new int[maxW];
    maxWindow = maxW;
    maxBytesWindow = bytesWindow;
    begin = 0;
    nextseqno = 0;    
    for (int i=0; i<maxW; i++) {
	sent[i] = 0;
	resent[i] = 0;
	acked[i] = 0;
    }
    bytesInTransit = 0;
    segmentsInTransit = 0;
}

AckScoreboard::~AckScoreboard()
{
    delete [] acked;
    delete [] sent;
    delete [] resent;
    delete [] msgSize;
}

void AckScoreboard::RecordMsg(int seqno, int size, int final, float t)
{
    sent[(begin + seqno - nextack) % maxWindow] = 1;
    resent[(begin + seqno - nextack) % maxWindow] = 0;    
    msgSize[(begin + seqno - nextack) % maxWindow] = size;
    fin[(begin + seqno - nextack) % maxWindow] = final;
    times[(begin + seqno - nextack) % maxWindow] = t;
    bytesInTransit += size;
    segmentsInTransit++;
}

void AckScoreboard::RecordResend(int seqno, float t)
{
    resent[(begin + seqno - nextack) % maxWindow] = 1;
    times[(begin + seqno - nextack) % maxWindow] = t;    
}

void AckScoreboard::RecordSendTime(int seqno, float t)
{
    times[(begin + seqno - nextack) % maxWindow] = t;    
}

void AckScoreboard::RecordAck(int seqno)
{
    if (seqno < nextack)
	return;
    /* old flow?? straggler packet?? */
    if (!sent[(begin + seqno - nextack) % maxWindow])
	return;
    if (acked[(begin + seqno - nextack) % maxWindow])
	return;
    bytesInTransit -= MsgSize(seqno);
    numNewAcks++;
    if (seqno == nextack) {
	acked[begin] = 1;
	while (acked[begin]) {
	    acked[begin] = 0;
	    begin = (begin + 1) % maxWindow;
	    nextack++;
	}
    } else
	acked[(begin + seqno - nextack) % maxWindow] = 1;
    segmentsInTransit--;
}

void AckScoreboard::RecordAck(int seqno, float timeSent)
{
    if (seqno < nextack)
	return;
    /* old flow?? straggler packet?? */
    if (!sent[(begin + seqno - nextack) % maxWindow])
	return;
    ackedTime[(begin + seqno - nextack) % maxWindow] = timeSent;
    if (acked[(begin + seqno - nextack) % maxWindow])
	return;
    bytesInTransit -= MsgSize(seqno);
    numNewAcks++;
    if (seqno == nextack) {
	acked[begin] = 1;
	while (acked[begin]) {
	    acked[begin] = 0;
	    begin = (begin + 1) % maxWindow;
	    nextack++;
	}
    } else
	acked[(begin + seqno - nextack) % maxWindow] = 1;
    segmentsInTransit--;
}

int AckScoreboard::CanSendMsg()
{
    /* actually a bit too optimistic in that it sends even if window could be exceeded */
    if ((nextseqno < nextack + maxWindow) && (bytesInTransit < maxBytesWindow))
	return 1;
    else
	return 0;
}

int AckScoreboard::CanSendNMsgs(int n)
{
    /* actually a bit too optimistic in that it sends even if window could be exceeded */
    if (((nextseqno + n - 1) < nextack + maxWindow) && (bytesInTransit < maxBytesWindow))
	return 1;
    else
	return 0;
}

int AckScoreboard::Receivedp(int seq)
{
    if (seq < nextack) return 1;
    if (acked[(begin + seq - nextack) % maxWindow])
	return 1;
    return 0;
}

float AckScoreboard::AckedTime(int seq)
{
    if (seq < nextack) return 0;
    return ackedTime[(begin + seq - nextack) % maxWindow];
}

int AckScoreboard::Resentp(int seq)
{
    assert (seq >= nextack);
    if (resent[(begin + seq - nextack) % maxWindow])
	return 1;
    return 0;
}

float AckScoreboard::TimeSent(int seq)
{
    assert (seq >= nextack);
    return times[(begin + seq - nextack) % maxWindow];
}

int AckScoreboard::MsgSize(int seqno)
{
    return msgSize[(begin + seqno - nextack) % maxWindow];
}

int AckScoreboard::MsgFinal(int seqno)
{
    return fin[(begin + seqno - nextack) % maxWindow];
}

int AckScoreboard::CanFastRetransmit(int seqno)
{
    if (Receivedp(seqno + 1) && Receivedp(seqno + 2) && Receivedp(seqno + 3))
	return 1;
    else
	return 0;
}

int AckScoreboard::CheckFastRetransmit()
{
    if (numNewAcks < 3)
	return -1;
    
    for (int i=nextack; i<nextseqno; i++) {
	if (Resentp(i))
	    return -1;
	if (!Receivedp(i)) {
	    int numDupacks = 0;
	    for (int j=i+1; j<nextseqno; j++)
		if (Receivedp(j))
		    numDupacks++;
	    if (numDupacks == 3) {
		numNewAcks = 0;
		return i;
	    }
	}
    }
    return -1;
}

int AckScoreboard::CheckRenoRetransmit()
{
    if (numNewAcks < 3)
	return -1;
    
    for (int i=nextack; i<nextseqno; i++) {
	if (!Receivedp(i) && !Resentp(i)) {
	    int numDupacks = 0;
	    for (int j=i+1; j<nextseqno; j++)
		if (Receivedp(j))
		    numDupacks++;
	    if (numDupacks == 3) {
		numNewAcks = 0;
		return i;
	    }
	}
    }
    return -1;
}

int AckScoreboard::CheckRetransmit()
{
    if ((nextseqno - nextack) < (segmentsInTransit + 3))
	return -1;
    
    for (int i=nextack; i<nextseqno; i++) {
	if (!Receivedp(i)) {
	    int numDupacks = 0;
	    for (int j=i+1; j<nextseqno; j++)
		if ((TimeSent(i) < TimeSent(j)) && (Receivedp(j))) {
		    if (TimeSent(i) < AckedTime(j)) {
			numDupacks++;
			if (numDupacks == 3)
			    return i;
		    }
		}
	}
    }
    return -1;
}
