#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "tscp.h"

/* #define DEBUG */

TcpSender::TcpSender(Network *n, int p, int pktSize, float think, TcpGlobalStats *stats)
{
    net = n;
    payload = p;
    packetSize = pktSize-40;
    protoOvhd = 40;
    conn = NULL;
    board = new AckScoreboard(800);
    flowNo = n->GetFlowNo();

    startTime = 0;
    finishTime = 0;
    thinkTime = think;
    isActive = 0;
    n->SetFlow(flowNo, this);
    globStats = stats;
}

void TcpSender::Start(float t)
{
    Event *e = new Event(t, this, START_TIMER, NULL, t);
    net->AddEvent(e);
}

void TcpSender::RepeatedStart(int num)
{
    int i;
    float t = 0;
    for (i=0; i<num; i++) {
	t += ThinkTimeDist(thinkTime);	
	/* t += thinkTime*(0.7 + 0.6*((random() % 100)/100.0)); */
	Event *e = new Event(t, this, START_TIMER, NULL, t);
	net->AddEvent(e);
    }
}

void TcpSender::CleanState()
{
    bytesSent = 0;
    rto = 3000; 	  /* 3secs - a large value */
    rttEstimate = 0;
    rttVarEstimate = 0;
    senderDone = 0;
    ssthresh = 200;
    cwin = 0;
    board->CleanScoreboard();
    inFastRetransmit = 0;
    reordSeq = -1;
}

void TcpSender::FinishFlow(float curr)
{
    UnregisterFlow();
    senderDone = 1;
    if (!isActive) {
	finishTime = startTime + (curr - startTime)*(payload/bytesSent);	
    } else {
	isActive = 0;
	finishTime = curr;
    }
    if (globStats)
	globStats->StopFlowStats(this);    
    TrackResponseTime(finishTime - startTime, bytesSent, 0, numPackets, numUniquePackets);
}

/* sends a message -- could be a retransmission */
void TcpSender::SendMessage(int seq, int size, float curr, int fin)
{
    Message *m = MakeMessage(flowNo, 0, size, conn, seq, fin, curr);
    m->conn->agents[0]->HandleMessage(m, curr);
    numPackets++;
#ifdef DEBUG
    printf("%f: message %d:%d (size %d) sent\n", curr, flowNo, seq, size);
#endif
#if 0
    Event *retEv = new Event(curr + RTO(), this, TIMEOUT, (void *)seq, curr);
    net->AddEvent(retEv);
    nextTimeout = curr + RTO();
#endif
    
}

void TcpSender::RecordedSendMessage(int seq, int size, float curr, int fin)
{
    if (size == 0 && seq != 0) return;
    SendMessage(seq, size + protoOvhd, curr, fin);
    bytesSent += size;
    board->RecordMsg(seq, size, fin, curr);
    board->AdvanceSeqNo();
    numUniquePackets++;    
}

void TcpSender::RegisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->RegisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->RegisterFlow(flowNo);
}

void TcpSender::UnregisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->UnregisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->UnregisterFlow(flowNo);
}

void TcpSender::HandleEvent(Event *e)
{
    int type = e->evType;
    float curr = e->time;

    if (type == START_TIMER) {
	/* interrupting a flow */
	if (isActive) {
	    FinishFlow(curr);
	}
	CleanState();
	startTime = curr;
	isActive = 1;
	RegisterFlow();
	RecordedSendMessage(0, 0, curr, 0);
	if (globStats)
	    globStats->StartFlowStats(this);	
	RechargeTimeout(curr);
	return;
    }

    if (senderDone) return;

    if (e->initTime < startTime) return;

    if (type == ENQUEUE_TIMER) {

	HandleMessage((Message *)(e->data), curr);

    } else if (type == TIMEOUT) {

	if (nextTimeout > curr)
	    return;
	assert(board->UnackedSegments() > 0);
	int pktSeqNo = board->OldestSeqNo();
	    
	/*	printf("timeout********** for %d:%d at %f\n", flowNo, pktSeqNo, curr); */
	SendMessage(pktSeqNo, board->MsgSize(pktSeqNo), curr, board->MsgFinal(pktSeqNo));
	rto = 2*rto;
	RechargeTimeout(curr);
	ssthresh = (cwin/2 > 1 ? cwin/2 : 1);
	cwin = 1;
    }
    
}

void TcpSender::RechargeTimeout(float curr)
{
    Event *retEv = new Event(curr + RTO(), this, TIMEOUT, 0, curr);
    net->AddEvent(retEv);
    nextTimeout = curr + RTO();
}

void TcpSender::HandleMessage(Message *m, float curr)
{
    if (board->Receivedp(m->seqno))
	return;

    if (m->ack2 != -1)
	board->RecordAck(m->ack2);
    board->RecordAck(m->seqno);
    
#ifdef DEBUG
    printf("%f: message %d:%d ack received, rtt: %f\n", curr, m->flowNo, m->seqno, curr - m->startTime);
#endif

    if (board->UnackedBytes() == 0 && payload == bytesSent) {
	FinishFlow(curr);
	return;
    }
    
    if (m->seqno == 0) {
	
	rttEstimate = curr - m->startTime;
	/* initialize rttVarEstimate to something high */
	rttVarEstimate = (curr - m->startTime)/2;
	rto = rttEstimate + 4*rttVarEstimate;
	
    } else {
	
	float newRtt = curr - m->startTime;
	float newRttVar = fabs(newRtt - rttEstimate);

	rttEstimate = 0.875*rttEstimate + 0.125*newRtt;
	rttVarEstimate = 0.75*rttVarEstimate + 0.25*newRttVar;
	rto = rttEstimate + 4*rttVarEstimate;
    }
    RechargeTimeout(curr);

    if (m->seqno == reordSeq) {
	inFastRetransmit = 0;
	reordSeq = -1;
    }
	
    if (!inFastRetransmit) {
	reordSeq = board->CheckFastRetransmit();
	if (reordSeq != -1) {
	    inFastRetransmit = 1;
	    ssthresh = (cwin/2 > 1 ? cwin/2 : 1);
	    cwin = (cwin/2 > 1 ? cwin/2 : 1);
	    /* perform a retransmission */
	    SendMessage(reordSeq, board->MsgSize(reordSeq), curr, board->MsgFinal(reordSeq));
	    board->RecordResend(reordSeq, curr);
	    return;
	} else {
	    if (cwin < ssthresh)
		cwin += 1;
	    else
		cwin += 1/cwin;
	}
    }
    

    int newpkts = (int)cwin - board->UnackedSegments();
    for (int i=0; i<newpkts; i++) {
	if (board->CanSendMsg()) {
	    int seq = board->NextSeqNo();
	    int sz = SendPacketSize();
	    if ((bytesSent + sz) == payload)
		RecordedSendMessage(seq, SendPacketSize(), curr + i*0.001, 1);
	    else
		RecordedSendMessage(seq, SendPacketSize(), curr + i*0.001, 0);
	}
    }
    
    
    delete m;
}

TcpGlobalStats::TcpGlobalStats(Network *n, int maxflows, float bneck)
{
    net = n;
    maxFlows = maxflows;
    numActiveFlows = 0;
    flows = new (TcpSender *)[maxFlows];
    for (int i=0; i<maxFlows; i++)
	flows[i] = NULL;
    bottleneck = bneck;
}

void TcpGlobalStats::StartFlowStats(TcpSender *s)
{
    int id = s->flowNo;
    assert(id < maxFlows);
    assert(flows[id] == NULL);
    flows[id] = s;
    numActiveFlows++;
}

void TcpGlobalStats::StopFlowStats(TcpSender *s)
{
    int id = s->flowNo;
    assert(id < maxFlows);
    assert(flows[id] != NULL);
    flows[id] = NULL;
    numActiveFlows--;
}

void TcpGlobalStats::StopAllFlows(float curr)
{
    for (int i=0; i<maxFlows; i++)
	if (flows[i])
	    flows[i]->FinishFlow(curr);
}
