#ifndef _TSCP_H
#define _TSCP_H

#include "stats.h"
#include "samples.h"

class Agent;

typedef struct {
    int numAgents;
    Agent **agents;
} Connection;

typedef struct {
    int flowNo;
    int msgType;	/* currently unused */
    int uniqueId;
    int size;
    Connection *conn;
    int seqno;
    int ack2;
    int fin;
    float bwshare;
    float qlen;
    int burstMsg;

    float startTime;
    float midTime;	/* when packet reaches dest */
    float endTime;	/* when ack comes back to src */
} Message;

Message *MakeMessage(int flowNo, int type, int size, Connection *conn,
		     int seqno, int fin, float time);
Connection *MakeConnection(int num, Agent **agents);
void TrackResponseTime(float, int bytes, int numRex, int numPackets, int numUniquePackets);
void EnableStats();
void PrintStats(float period);
float ThinkTimeDist(float mean);
float ExpTimeDist(float mean);

#include "events.h"

class Event {
 public:
    Event(float t, Agent *a, int type, void *data, float curr);
    float time;
    Agent *agent;
    int evType;
    void *data;
    float initTime;
};

#include "heap.h"

typedef struct HistoryRecordStruct {
    Agent *s, *d;
    float bw;
    float minrtt;
    HistoryRecordStruct *next;
} HistoryRecord;

class LinkAgent;
class SenderAgent;
class TscpGlobalStats;
class TcpGlobalStats;

class Network {
 public:
    Network();
    void HandleEvent(Event *);
    void AddEvent(Event *);
    int Step();
    int GetFlowNo();
    int GetLinkNo();
    float GetTime() { return currTime; }
    void RecordHistory(Agent *s, Agent *d, float bw, float minrtt);
    float LookupHistory(Agent *s, Agent *d, float *minrtt);
    void SetLink(int linkno, LinkAgent *l) {
	assert(linkno < numLinks);
	links[linkno] = l;
    }
    void SetFlow(int flowno, SenderAgent *s) {
	assert(flowno < numFlows);
	senders[flowno] = s;
    }
    void SetTscpStats(TscpGlobalStats *s) { tscpStats = s; }
    void SetTcpStats(TcpGlobalStats *s) { tcpStats = s; }
    void CleanStats();
    
 private:
    float currTime;
    Heap *events;
    int maxFlows;
    int maxLinks;
    int numFlows;
    int numLinks;
    HistoryRecord *history;
    LinkAgent **links;
    SenderAgent **senders;
    TscpGlobalStats *tscpStats;
    TcpGlobalStats *tcpStats;    
};

class Agent {
 public:
    virtual void HandleMessage(Message *m, float currTime) = 0;
    virtual void HandleEvent(Event *) = 0;
};

class LinkAgent : public Agent {
 public:
    virtual void HandleMessage(Message *m, float currTime) = 0;
    virtual void HandleEvent(Event *) = 0;
    virtual void RegisterFlow(int flowId) = 0;
    virtual void UnregisterFlow(int flowId) = 0;
    virtual void EnableStats(float) = 0;
    virtual void PrintStats(float) = 0;
    virtual int LinkId() = 0;
};

class FifoLink : public LinkAgent {
 public:
    FifoLink(Network *, int max_msgs, float bw, float del, int mon);
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
    void RegisterFlow(int flowId) { }
    void UnregisterFlow(int flowId) { }
    void EnableStats(float currTime);
    void PrintStats(float currTime);
    void UpdateQueueStats(float currTime);
    int LinkId() { return linkId; }
    float Delay() { return delay; }
    float Bandwidth() { return bandwidth; }
 private:
    Network *net;
    int maxMessages;
    int begin, end, numMessages;
    int numBytes, maxBytes, numBurstBytes;
    float bandwidth;
    float delay;
    Message **buffer;
    int linkId;

    int statsNumBytes;
    int statsNumDrops;
    int statsNumPackets;
    float statsQueueSize;
    float statsBurstQueueSize;
    float statsLastEvent;
    float statsInitTime;

    StatsObject *statsObj;
    int statsMon;
};

class FQFlow {
 public:
    void InitFlow(int max_msgs, int segSize);
    void StartFlow(int fid);
    int IsDormant();
    void EndFlow(Heap *szHeap, Heap *priorHeap, Heap *actHeap, FQFlow **xmitFlow);
    int AddTail(Message *m, Heap *);
    Message *RemoveHead(Heap *, Heap *);
    Message *PeepHead();
    Message *RemoveTail();
    int NumMessages() { return numMessages; }
    void ResizeFlow(Heap *);
    void AddToSizeHeap(Heap *);
    void AddToActiveHeap(Heap *);
    void SetStartRound(float newRound, Heap *priorHeap);
    void PrintStatus();
    void SetActive() { inActiveHeap = 1; }
    void UnsetActive() { inActiveHeap = 0; }
    int InActiveHeap() { return inActiveHeap; }
    float GetBWShare() { return allocShare; }
    void SetBWShare(float bw) { allocShare = bw; }
    
 private:
    int flowId;
    int begin, end, numMessages, maxMessages;
    int numBytes;
    Message **buffer;
    int segmentSize;
    float startRound;
    float finishRound;
    float allocShare;

    int sizeHeapPos;
    int priorHeapPos;
    int activeHeapPos;
    int inActiveHeap;
};

class FQLink : public LinkAgent {
 public:
    FQLink(Network *, int max_msgs, int pktSize,
	   float bw, float del, int maxFlows, int maxFlowId, int mon);
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
    void RegisterFlow(int flowId);
    void UnregisterFlow(int flowId);
    void UpdateVirtualTime(float currTime);
    void PrintStatus();
    void EnableStats(float currTime);
    void PrintStats(float currTime);
    void UpdateQueueStats(float currTime);
    int LinkId() { return linkId; }
 private:
    Network *net;
    int maxFlows;
    FQFlow *flows;
    FQFlow *xmitFlow;
    
    int maxFlowId;
    int *invFlowIdMap;

    int numMessages;
    int maxMessages;
    Heap *flowPriority;
    Heap *flowSize;
    
    float bandwidth;
    float delay;
    int linkId;
    int segmentSize;

    float virtRound;
    float virtTime;
    int numActiveFlows;
    int numRegisteredFlows;
    
    Heap *sizeHeap;
    Heap *priorHeap;
    Heap *stillActiveHeap;


    int statsNumBytes;
    int statsNumDrops;
    int statsNumPackets;
    float statsQueueSize;
    float statsLastEvent;
    float statsInitTime;

    StatsObject *statsObj;
    int statsMon;
};

class AckScoreboard {
 public:
    AckScoreboard(int maxW);
    AckScoreboard(int maxW, int bytesWindow);    
    ~AckScoreboard();
    void CleanScoreboard();
    void RecordMsg(int seqno, int size, int fin, float t);
    void RecordResend(int seqno, float t);
    void RecordSendTime(int seqno, float t);
    void RecordAck(int seqno);
    void RecordAck(int seqno, float timeSent);
    int CanSendMsg();
    int CanSendNMsgs(int n);
    int NextSeqNo() { return nextseqno; }
    int OldestSeqNo() { return nextack; }
    void AdvanceSeqNo() { nextseqno++; }
    int Receivedp(int);
    float AckedTime(int);
    int Resentp(int);
    float TimeSent(int seq);
    int MsgSize(int seqno);
    int MsgFinal(int);
    void SetBytesWindow(int sz) { maxBytesWindow = sz; }
    int UnackedBytes() { return bytesInTransit; }
    int UnackedSegments() { return segmentsInTransit; }
    int CanFastRetransmit(int);
    int CheckFastRetransmit();
    int CheckRenoRetransmit();
    int CheckRetransmit();    
 private:
    int nextack;
    int *sent;
    int *resent;
    int *acked;
    float *ackedTime;
    int *msgSize;
    int *fin;
    float *times;
    int maxWindow;
    int begin;
    int nextseqno;
    int maxBytesWindow;
    int bytesInTransit;
    int segmentsInTransit;
    int numNewAcks;
};

class SenderAgent : public Agent {
 public:
    virtual void HandleMessage(Message *m, float currTime) = 0;
    virtual void HandleEvent(Event *) = 0;
    virtual void SetConnection(Connection *c) = 0;
    virtual void Start(float) = 0;
    virtual void RepeatedStart(int) = 0;
};

class TscpGlobalStats;

class TscpSender: public SenderAgent {
    friend class TscpGlobalStats;
 public:
    TscpSender(Network *, int payload, int pktSize, float thinkTime, TscpGlobalStats *stats);
    /*    TscpSender(int payload, Connection *c); */
    void SetConnection(Connection *c) { conn = c; }
    void Start(float);
    void RepeatedStart(int);
    void CleanState();
    void FinishFlow(float curr);
    
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
    void RegisterFlow();
    void UnregisterFlow();

    void RegisterDrop(int);
    int CheckDropQueue();
    int InDropQueue(int seq);

 private:

    void RecordedSendMessage(int seq, int size, float curr, int burst);
    void SendMessage(int seq, int size, float curr, int burst);

    void UpdateUnderlyingRate(float newRate, float curr);
    void UCompensate(float curr);

    void StartNewBurst(float curr);
    void SendBurstMessage(float curr);
    void ProcessBurstResults(float curr);
    void DebugCheckBurstFailure();    
    int CheckBurstFailure();
    void RestartBurst(int pktSeq, float curr);
    void RechargeTimeout(float curr);
    
    float RTO() { return 1.5*rttEstimate + 4*rttVarEstimate; }
    float SendInterval() { return (packetSize + protoOvhd)/uniformRate; }
    int SendPacketSize() {
	return (((payload - bytesSent) > packetSize)
		? packetSize : (payload - bytesSent));
    }

    Network *net;
    Connection *conn;
    int flowNo;
    int payload;
    int bytesSent;
    int senderDone;
    int isActive;
    
    int packetSize;
    int protoOvhd;
    AckScoreboard *board;
    int *dropSeq;
    int numDrops;
    int numRexmits;
    int numPackets;
    int numUniquePackets;

    float rttEstimate;
    float rttVarEstimate;
    float minRtt;

    float uniformRate;
    float lastXmitTime, nextXmitTime;
    float maxRtt;
    float nextTimeout;

    int burstMsgsSent;
    int burstMsgsRcvd;
    int burstMsgSeqno[5];
    float burstMsgRtt[5];
    float burstStartTime;
    int burstPhase;
    float burstSpacing;
    int burstMsgSize;
    float currBurstRate;
    

    float startTime;
    float finishTime;
    float thinkTime;

    float compQmax;
    float compTime;

    TscpGlobalStats *globStats;
    Samples *msgSizes;
    Samples *uRates;
    Samples *qSizes;
};

class GenericReceiver: public Agent {
 public:
    GenericReceiver(Network *);
    GenericReceiver(Network *, Connection *);
    void SetConnection(Connection *c) { conn = c; }
    virtual void HandleMessage(Message *m, float currTime);
    virtual void HandleEvent(Event *);
    void RegisterFlow(int);
    void UnregisterFlow(int);
    
 protected:
    Network *net;
    Connection *conn;
};

class TcpReceiver: public GenericReceiver {
 public:
    TcpReceiver(Network *);
    TcpReceiver(Network *, Connection *);
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
 private:
    int rememberedAck;
    int *acked;
    int maxWindow;
    int begin;
    int nextAck;
    int numOutOfOrderAcks;
};

class NullAgent: public Agent {
 public:
    NullAgent() { }
    void HandleMessage(Message *m, float currTime) {
	printf("Time: %f\n", currTime);
    }
    void HandleEvent(Event *e) {
	printf("Time: %f\n", e->time);
    }
};

class TcpSender: public SenderAgent {
    friend class TcpGlobalStats;    
 public:
    TcpSender(Network *, int payload, int pktSize, float think, TcpGlobalStats *t);
    void SetConnection(Connection *c) { conn = c; }
    void Start(float);
    void RepeatedStart(int);

    void CleanState();
    void FinishFlow(float curr);
    
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
    void RechargeTimeout(float curr);
    void RegisterFlow();
    void UnregisterFlow();
    
 private:

    void RecordedSendMessage(int seq, int size, float curr, int fin);
    void SendMessage(int seq, int size, float curr, int fin);

    float RTO() { return rto; }
    int SendPacketSize() {
	return (((payload - bytesSent) > packetSize)
		? packetSize : (payload - bytesSent));
    }

    Network *net;
    Connection *conn;
    int flowNo;
    int payload;
    int bytesSent;
    int senderDone;
    int isActive;

    int packetSize;
    int protoOvhd;
    AckScoreboard *board;
    int numPackets;
    int numUniquePackets;

    float ssthresh;
    float cwin;
    int inFastRetransmit;
    int reordSeq;

    float rttEstimate;
    float rttVarEstimate;
    float rto;
    float minRtt;
    float nextTimeout;

    float startTime;
    float finishTime;
    float thinkTime;

    TcpGlobalStats *globStats;    
};


class FQSender: public SenderAgent {
 public:
    FQSender(Network *, int payload, int pktSize, float think);
    void SetConnection(Connection *c) { conn = c; }
    void Start(float);
    void RepeatedStart(int);
    void CleanState();
    void FinishFlow(float);
    
    void HandleMessage(Message *m, float currTime);
    void HandleEvent(Event *);
    void RechargeTimeout(float curr);
    void RegisterFlow();
    void UnregisterFlow();
    void UpdateSpacing(float newspacing, float curr);
    void RecordSample(int seq, float rtt, float curr, float bwshare, float qlen);
    
 private:

    void RecordedSendMessage(int seq, int size, float curr);
    void SendMessage(int seq, int size, float curr);

    float RTO() { return 2*rto; }
    int SendPacketSize() {
	return (((payload - bytesSent) > packetSize)
		? packetSize : (payload - bytesSent));
    }

    Network *net;
    Connection *conn;
    int flowNo;
    int payload;
    int bytesSent;
    int senderDone;
    int isActive;
    int numRexmits;
    int numPackets;
    int numUniquePackets;

    int packetSize;
    int protoOvhd;
    AckScoreboard *board;

    float spacing;
    float nextXmit;
    float lastXmit;

    /* state to remember the rtts of a packet pair */
    int numSamples;
    int maxSamples;
    int *seqNos;
    float *msgInitTimes;
    float *rttSamples;

    float rttEstimate;
    float rttVarEstimate;
    float minRtt;
    float rto;
    float nextTimeout;
    float myShare;
    float qLen;

    float startTime;
    float finishTime;
    float thinkTime;    
};

class TscpGlobalStats {
 public:
    TscpGlobalStats(Network *n, int maxflows, float bneck);
    void StopAllFlows(float curr);
    void StartFlowStats(TscpSender *);
    void StopFlowStats(TscpSender *);
    void UpdateURateStats(float oldRate, float newRate, float curr);
    void StartBurstStats(float curr, float rate);
    void StopBurstStats(float curr, float rate);
    void RecordBurstSuccess(float rate, int success, int cancel, float btime);
    void PrintCumStats();
    void CleanStats();
    void PrintStats();
    void PrintAllFlows(float);
    void TrackBPLoss() { burstPacketloss++; }
 private:
    Network *net;
    int maxFlows;
    int numActiveFlows;
    TscpSender **flows;
    float cumURate;
    float cumBRate;
    float bottleneck;
    int numBursting;

    float *recentUrates;
    float *recentTimes;
    
    float *recentBrates;
    float *recentBTimes;
    int nextSample;
    int nextBSample;
    int maxSamples;

    int *sprobeBWBins;
    int *fprobeBWBins;
    int *fnprobeBWBins;
    int *fpprobeBWBins;
    
    int falsePositives, falseNegatives;
    int burstSuccesses, burstFailures;
    int burstPacketloss;
};

class TcpGlobalStats {
 public:
    TcpGlobalStats(Network *n, int maxflows, float bneck);
    void StopAllFlows(float curr);
    void StartFlowStats(TcpSender *);
    void StopFlowStats(TcpSender *);
 private:
    Network *net;
    int maxFlows;
    int numActiveFlows;
    TcpSender **flows;
    float bottleneck;
};
#endif
